function [primal, dual, how] = optimize_smw(optimizer, c, Hmn, Hnn, A, b, r, l, u)
%[PRIMAL, DUAL, HOW] = OPTIMIZE_SMW(optimizer, c, Hmn, Hnn, A, b, r, l, u)
%
%optimize solves the quadratic programming problem
%
%minimize   c' * primal + 1/2 primal' * H * primal
%subject to b <= A*primal <= b + r
%           l <= x <= u
%           d is the optimizer itself
%
%returns primal and dual variables (i.e. x and the Lagrange
%multipliers for b <= A * primal <= b + r
%
%for additional documentation see 
%     R. Vanderbei 
%     LOQO: an Interior Point Code for Quadratic Programming, 1992

% File:        @intpoint/optimize_smw.m
%
% Author:      Alex J. Smola
% Created:     12/12/97
% Updated:     08/05/00
%
% This code is released under the GNU Public License
% Copyright by GMD FIRST and The Australian National University

% gather some starting data
[m, n] = size(A);
n2 = size(Hmn, 2);

b_plus_1 = norm(b) + 1;
c_plus_1 = norm(c) + 1;
one_x = -ones(n,1);
one_y = -ones(m,1);

% starting point
H_y = eye(m);
c_x = c;
c_y = b;

smw_step2 = inv(Hnn + Hmn' * Hmn); 
% worth doing it since we have to use it several times
A_Hmn = A * Hmn;
smw_H_tmp = A * A' - A_Hmn * (smw_step2 * A_Hmn') + H_y;
y = smw_H_tmp \ (c_y + A * (c_x - (Hmn * (smw_step2 * (Hmn' * c_x))));
smw_step6 = A' * y - c_x; % D=1: no scaling
x = smw_step6 - (Hmn * (smw_step2 * (Hmn' * smw_step6)));

g = max(abs(x - l), optimizer.bound);
z = max(abs(x), optimizer.bound);
t = max(abs(u - x), optimizer.bound);
s = max(abs(x), optimizer.bound);
v = max(abs(y), optimizer.bound);
w = max(abs(y), optimizer.bound);
p = max(abs(r - w), optimizer.bound);
q = max(abs(y), optimizer.bound);

mu = (z' * g + v' * w + s' * t + p' * q)/(2 * (m + n));

sigfig = 0;

counter = 0;

alfa = 1;

if (optimizer.verbose > 0)	% print at least one status report
  report = sprintf(['Iter \tPrimalInf \tDualInf \tSigFigs ' ...
		    '\tRescale \tPrimalObj \tDualObj']); 
  disp(report);
end

%H_full = Hmn * inv(Hnn) * Hmn';

while (counter < optimizer.maxiter),
  %update the iteration counter
  counter = counter + 1;
  
  %central path (predictor)
  %H_dot_x = H_full * x;
  H_dot_x = (Hmn * (inv(Hnn) * (Hmn' * x)));
  
  rho = b - A * x + w;
  nu = l - x + g;
  tau = u - x - t;
  alpha = r - w - p;
  
  sigma = c - A' * y - z + s + H_dot_x;
  beta = y + q - v;
  
  gamma_z = - z;
  gamma_w = - w;
  gamma_s = - s;
  gamma_q = - q;
  
  % instrumentation
  x_dot_H_dot_x = x' * H_dot_x;
  
  primal_infeasibility = norm([rho; tau; alpha; nu]) / b_plus_1;
  dual_infeasibility = norm([sigma; beta]) / c_plus_1;
  
  primal_obj = c' * x + 0.5 * x_dot_H_dot_x;
  dual_obj = b' * y - 0.5 * x_dot_H_dot_x + l' * z - u' * s - r' * q;
  
  old_sigfig = sigfig;
  sigfig = max(-log10(abs(primal_obj - dual_obj)/(abs(primal_obj) + 1)), 0);
  
  if (sigfig >= optimizer.sigfig) break; end;
  if (optimizer.verbose > 1)			% final report
    report = sprintf('%i \t%e \t%e \t%e \t%e \t%e \t%e \t%e', ...
		     counter, primal_infeasibility, dual_infeasibility, ...
		     sigfig, alfa, primal_obj, dual_obj, sum(c_plus_1));
    disp(report);
  end
  
  % some more intermediate variables (the hat section)
  hat_beta = beta - v .* gamma_w ./ w;
  hat_alpha = alpha - p .* gamma_q ./ q;
  hat_nu = nu + g .* gamma_z ./ z;
  hat_tau = tau - t .* gamma_s ./ s;
  
  % the diagonal terms
  d = z ./ g + s ./ t;
  e = 1 ./ (v ./ w + q ./ p);
  
  % initialization before the big cholesky
  H_y = diag(e);
  c_x = sigma - z .* hat_nu ./ g - s .* hat_tau ./ t;
  c_y = rho - e .* (hat_beta - q .* hat_alpha ./ p);

  % Do the same thing using S-M-W.
  for i=1:n2
    smw_step1(:,i) = Hmn(:,i) ./ d;
  end;		   
  smw_step2 = inv(Hnn + Hmn' * smw_step1);

  % We will use this later
  for i=1:m
    smw_step3(:,i) = (A(i,:) ./ d')';
  end
  smw_step4 = smw_step3 - (smw_step1 * (smw_step2 * (Hmn' * smw_step3)));
  smw_H_tmp = A * smw_step4 + H_y;
  smw_inv_H_tmp = inv(smw_H_tmp);
  for i=1:m
    smw_step5(:,i) = c_x(:,i) ./ d;
  end
  delta_y = smw_inv_H_tmp * (c_y + A * (smw_step5 - (smw_step1 * ...
						  (smw_step2 * (Hmn' * smw_step5)))));
  smw_step6_t = A' * delta_y - c_x;
  for i=1:m
    smw_step6(:,i) = smw_step6_t(:,i) ./ d;
  end
  delta_x = smw_step6 - (smw_step1 * (smw_step2 * (Hmn' * smw_step6)));  
  
  %backsubstitution
  delta_w = - e .* (hat_beta - q .* hat_alpha ./ p + delta_y);
  
  delta_s = s .* (delta_x - hat_tau) ./ t;
  delta_z = z .* (hat_nu - delta_x) ./ g;
  delta_q = q .* (delta_w - hat_alpha) ./ p;
  
  delta_v = v .* (gamma_w - delta_w) ./ w;
  delta_p = p .* (gamma_q - delta_q) ./ q;
  delta_g = g .* (gamma_z - delta_z) ./ z;
  delta_t = t .* (gamma_s - delta_s) ./ s;
  
  %compute update step now (sebastian's trick)
  alfa = - (1 - optimizer.margin) / min([delta_g ./ g; delta_w ./ w; delta_t ./ t;
                    delta_p ./ p; delta_z ./ z; delta_v ./ v;
		    delta_s ./ s; delta_q ./ q; -1]);
  
  newmu = (z' * g + v' * w + s' * t + p' * q)/(2 * (m + n));
  newmu = mu * ((alfa - 1) / (alfa + 10))^2;

  %central path (corrector)
  gamma_z = mu ./ g - z - delta_z .* delta_g ./ g;
  gamma_w = mu ./ v - w - delta_w .* delta_v ./ v;
  gamma_s = mu ./ t - s - delta_s .* delta_t ./ t;
  gamma_q = mu ./ p - q - delta_q .* delta_p ./ p;
  
  % some more intermediate variables (the hat section)
  hat_beta = beta - v .* gamma_w ./ w;
  hat_alpha = alpha - p .* gamma_q ./ q;
  hat_nu = nu + g .* gamma_z ./ z;
  hat_tau = tau - t .* gamma_s ./ s;
  
  % the diagonal terms
  %d = z ./ g + s ./ t;
  %e = 1 ./ (v ./ w + q ./ p);
  
  % initialization before the big cholesky
  %for i = 1:n H_x(i,i) = H_diag(i) + d(i);
  %H_y = diag(e);
  c_x = sigma - z .* hat_nu ./ g - s .* hat_tau ./ t;
  c_y = rho - e .* (hat_beta - q .* hat_alpha ./ p);
  
  % Note that D hasn't changed so we don't need to modify a whole bunch of things.
  for i=1:m
    smw_step5(:,i) = c_x(:,i) ./ d;
  end
  delta_y = smw_inv_H_tmp * (c_y + A * (smw_step5 - (smw_step1 * ...
						  (smw_step2 * (Hmn' ...
						  * smw_step5)))));
  smw_step6_t = A' * delta_y - c_x;
  for i=1:m
    smw_step6(:,i) = smw_step6_t(:,i) ./ d;
  end
  delta_x = smw_step6 - (smw_step1 * (smw_step2 * (Hmn' * smw_step6)));
  
  %backsubstitution
  delta_w = - e .* (hat_beta - q .* hat_alpha ./ p + delta_y);
  
  delta_s = s .* (delta_x - hat_tau) ./ t;
  delta_z = z .* (hat_nu - delta_x) ./ g;
  delta_q = q .* (delta_w - hat_alpha) ./ p;
  
  delta_v = v .* (gamma_w - delta_w) ./ w;
  delta_p = p .* (gamma_q - delta_q) ./ q;
  delta_g = g .* (gamma_z - delta_z) ./ z;
  delta_t = t .* (gamma_s - delta_s) ./ s;
  
  %compute the updates
  alfa = - (1 - optimizer.margin) / min([delta_g ./ g; delta_w ./ w; delta_t ./ t;
                    delta_p ./ p; delta_z ./ z; delta_v ./ v;
		    delta_s ./ s; delta_q ./ q; -1]);
  
% $$$   mu = (z' * g + v' * w + s' * t + p' * q)/(2 * (m + n));
% $$$   mu = mu * ((alfa - 1) / (alfa + 10))^2;

  mu = newmu;
  
  x = x + delta_x * alfa;
  g = g + delta_g * alfa;
  w = w + delta_w * alfa;
  t = t + delta_t * alfa;
  p = p + delta_p * alfa;
  y = y + delta_y * alfa;
  z = z + delta_z * alfa;
  v = v + delta_v * alfa;
  s = s + delta_s * alfa;
  q = q + delta_q * alfa;
end

if (optimizer.verbose > 0)			% final report
  report = sprintf('%i \t%e \t%e \t%e \t%e \t%e \t%e', ...
		   counter, primal_infeasibility, dual_infeasibility, ...
		   sigfig, alfa, primal_obj, dual_obj);    
  disp(report);
end

% repackage the results
primal = x;
dual   = y;

if ((sigfig > optimizer.sigfig) & (counter < optimizer.maxiter))
  how    = 'converged';
else					% must have run out of counts
  if ((primal_infeasibility > 10e5) & (dual_infeasibility > 10e5))
    how    = 'primal and dual infeasible';
  elseif (primal_infeasibility > 10e5)
    how    = 'primal infeasible';
  elseif (dual_infeasibility > 10e5)
    how    = 'dual infeasible';
  else					% don't really know
    how    = 'slow convergence, change bound?';
  end;
end









