function [primal, dual, how] = optimize_smw(optimizer, c, Hmn, Hnn, A, b, l, u)
%[PRIMAL, DUAL, HOW] = OPTIMIZE_SMW(optimizer, c, Hmn, Hnn, A, b, l, u)
%
%optimize solves the quadratic programming problem
%
%minimize   c' * primal + 1/2 primal' * H * primal
%subject to A*primal = b
%           l <= primal <= u
%
%returns primal and dual variables (i.e. primal and the Lagrange
%multipliers for A * primal <= b
%
% In this case H is approximated by the product Hmn * inv(Hnn) * Hmn'.
%
%for additional documentation see 
%     R. Vanderbei 
%     LOQO: an Interior Point Code for Quadratic Programming, 1992

% File:        @intpoint_pr/optimize_smw.m
%
% Author:      James Macnicol (modified from @intpoint_pr/optimize.m by Alex Smola)
% Created:     23/10/2000
%
% This code is released under the GNU Public License
% Copyright by GMD FIRST and The Australian National University

% gather some starting data
[m, n] = size(A); 
n2 = size(Hmn, 2);

b_plus_1 = 1;
c_plus_1 = norm(c) + 1;
one_x = -ones(n,1);
one_y = -ones(m,1);

% starting point
H_y = eye(m);
c_x = c;
c_y = b;

smw_step1 = Hmn;  % D=1 so the scaling here has no effect
smw_step2 = inv(Hnn + Hmn' * smw_step1);
smw_step3 = A'; % again, D=1 so no scaling
smw_step4 = smw_step3 - (smw_step1 * (smw_step2 * (Hmn' * smw_step3)));
smw_H_tmp = A * smw_step4 + H_y;
smw_step5 = c_x; % D=1: no scaling
y = inv(smw_H_tmp) * (c_y + A * (smw_step5 - (smw_step1 * (smw_step2 ...
						  * (Hmn' * smw_step5)))));
smw_step6 = A' * y - c_x; % D=1: no scaling
x = smw_step6 - (smw_step1 * (smw_step2 * (Hmn' * smw_step6)));

g = max(abs(x - l), optimizer.bound);
z = max(abs(x), optimizer.bound);
t = max(abs(u - x), optimizer.bound);
s = max(abs(x), optimizer.bound);

mu = (z' * g + s' * t)/(2 * n);

% set some default values
sigfig = 0;				% number of significant figures
counter = 0;				% iteration counter
alfa = 1;				% scaling factor

if (optimizer.verbose > 0)	% print at least one status report
  report = sprintf(['Iter \tPrimalInf \tDualInf \tSigFigs ' ...
		    '\tRescale \tPrimalObj \tDualObj']); 
  disp(report);
end

%H_full = Hmn * inv(Hnn) * Hmn';

while (counter < optimizer.maxiter),

  %update the iteration counter
  counter = counter + 1;
  
  %H_dot_x = H_full * x;
  H_dot_x = (Hmn * (inv(Hnn) * (Hmn' * x)));
  
  rho = - A * x + b;
  nu = l - x + g;
  tau = u - x - t;
  sigma = c - A' * y - z + s + H_dot_x;
  
  gamma_z = - z;
  gamma_s = - s;
  
  % instrumentation
  x_dot_H_dot_x = x' * H_dot_x;
  
  primal_infeasibility = norm([tau; nu; rho]) / b_plus_1;
  dual_infeasibility = norm([sigma]) / c_plus_1;
  
  primal_obj = c' * x + 0.5 * x_dot_H_dot_x;
  dual_obj = b' * y - 0.5 * x_dot_H_dot_x + l' * z - u' * s;
  
  old_sigfig = sigfig;
  sigfig = max(-log10(abs(primal_obj - dual_obj)/(abs(primal_obj) + 1)), 0);
  
  if (sigfig >= optimizer.sigfig) break; end;
  if (optimizer.verbose > 1)			% final report
    report = sprintf('%i \t%e \t%e \t%e \t%e \t%e \t%e', ...
		     counter, primal_infeasibility, dual_infeasibility, ...
		     sigfig, alfa, primal_obj, dual_obj);    
    disp(report);
  end

  % some more intermediate variables (the hat section)
  hat_nu = nu + g .* gamma_z ./ z;
  hat_tau = tau - t .* gamma_s ./ s;
  
  % the diagonal terms
  d = z ./ g + s ./ t;
  
  H_y = 0;
  c_x = sigma - z .* hat_nu ./ g - s .* hat_tau ./ t;
  c_y = rho;
  
  % Do the same thing using S-M-W.
  for i=1:n2
    smw_step1(:,i) = Hmn(:,i) ./ d;
  end;		   
  smw_step2 = inv(Hnn + Hmn' * smw_step1);

  % We will use this later
  for i=1:m
    smw_step3(:,i) = (A(i,:) ./ d')';
  end
  smw_step4 = smw_step3 - (smw_step1 * (smw_step2 * (Hmn' * smw_step3)));
  smw_H_tmp = A * smw_step4 + H_y;
  smw_inv_H_tmp = inv(smw_H_tmp);
  for i=1:m
    smw_step5(:,i) = c_x(:,i) ./ d;
  end
  delta_y = smw_inv_H_tmp * (c_y + A * (smw_step5 - (smw_step1 * ...
						  (smw_step2 * (Hmn' * smw_step5)))));
  smw_step6_t = A' * delta_y - c_x;
  for i=1:m
    smw_step6(:,i) = smw_step6_t(:,i) ./ d;
  end
  delta_x = smw_step6 - (smw_step1 * (smw_step2 * (Hmn' * smw_step6)));

  %backsubstitution
  delta_s = s .* (delta_x - hat_tau) ./ t;
  delta_z = z .* (hat_nu - delta_x) ./ g;
  
  delta_g = g .* (gamma_z - delta_z) ./ z;
  delta_t = t .* (gamma_s - delta_s) ./ s;
  
  %compute update step now (sebastian's trick)
  alfa = - (1-optimizer.margin) / min([delta_g ./ g; delta_t ./ t;
                    delta_z ./ z; delta_s ./ s; -1]);
  newmu = (z' * g + s' * t)/(2 * n);
  newmu = mu * ((alfa - 1) / (alfa + 10))^2;  
  
  %central path (corrector)
  gamma_z = mu ./ g - z - delta_z .* delta_g ./ g;
  gamma_s = mu ./ t - s - delta_s .* delta_t ./ t;
  
  % some more intermediate variables (the hat section)
  hat_nu = nu + g .* gamma_z ./ z;
  hat_tau = tau - t .* gamma_s ./ s;
  
  c_x = sigma - z .* hat_nu ./ g - s .* hat_tau ./ t;
  c_y = rho;
  
  % Note that D hasn't changed so we don't need to modify a whole bunch of things.
  for i=1:m
    smw_step5(:,i) = c_x(:,i) ./ d;
  end
  delta_y = smw_inv_H_tmp * (c_y + A * (smw_step5 - (smw_step1 * ...
						  (smw_step2 * (Hmn' ...
						  * smw_step5)))));
  smw_step6_t = A' * delta_y - c_x;
  for i=1:m
    smw_step6(:,i) = smw_step6_t(:,i) ./ d;
  end
  delta_x = smw_step6 - (smw_step1 * (smw_step2 * (Hmn' * smw_step6)));

  %backsubstitution
  
  delta_s = s .* (delta_x - hat_tau) ./ t;
  delta_z = z .* (hat_nu - delta_x) ./ g;
  
  delta_g = g .* (gamma_z - delta_z) ./ z;
  delta_t = t .* (gamma_s - delta_s) ./ s;
  
  %compute the updates
  alfa = - 0.95 / min([delta_g ./ g; delta_t ./ t;
                    delta_z ./ z; delta_s ./ s; -1]);
  
% $$$   mu = (z' * g + s' * t)/(2 * n);
% $$$   mu = mu * ((alfa - 1) / (alfa + 10))^2;
  mu = newmu;
  
  x = x + delta_x * alfa;
  g = g + delta_g * alfa;
  t = t + delta_t * alfa;
  y = y + delta_y * alfa;
  z = z + delta_z * alfa;
  s = s + delta_s * alfa;
  
end

if (optimizer.verbose > 0)			% final report
  report = sprintf('%i \t%e \t%e \t%e \t%e \t%e \t%e', ...
		   counter, primal_infeasibility, dual_infeasibility, ...
		   sigfig, alfa, primal_obj, dual_obj);    
  disp(report);
end

% repackage the results
primal = x;
dual   = y;

if ((sigfig > optimizer.sigfig) & (counter < optimizer.maxiter))
  how    = 'converged';
else					% must have run out of counts
  if ((primal_infeasibility > 10e5) & (dual_infeasibility > 10e5))
    how    = 'primal and dual infeasible';
  elseif (primal_infeasibility > 10e5)
    how    = 'primal infeasible';
  elseif (dual_infeasibility > 10e5)
    how    = 'dual infeasible';
  else					% don't really know
    how    = 'slow convergence, change bound?';
  end;
end



